import pytest

from torch_so3.angular_ranges import SymmetryRanges, get_symmetry_ranges


def test_default_symmetry():
    result = get_symmetry_ranges()
    expected = SymmetryRanges(
        phi_min=0.0,
        phi_max=360.0,
        theta_min=0.0,
        theta_max=180.0,
        psi_min=0.0,
        psi_max=360.0,
    )
    assert result == expected


def test_symmetry_group_C():
    result = get_symmetry_ranges(symmetry_group="C", symmetry_order=2)
    expected = SymmetryRanges(
        phi_min=0.0,
        phi_max=180.0,
        theta_min=0.0,
        theta_max=180.0,
        psi_min=0.0,
        psi_max=360.0,
    )
    assert result == expected


def test_symmetry_group_D():
    result = get_symmetry_ranges(symmetry_group="D", symmetry_order=2)
    expected = SymmetryRanges(
        phi_min=0.0,
        phi_max=180.0,
        theta_min=0.0,
        theta_max=90.0,
        psi_min=0.0,
        psi_max=360.0,
    )
    assert result == expected


def test_symmetry_group_T():
    result = get_symmetry_ranges(symmetry_group="T")
    expected = SymmetryRanges(
        phi_min=0.0,
        phi_max=180.0,
        theta_min=0,
        theta_max=54.7356,
        psi_min=0.0,
        psi_max=360.0,
    )
    assert result == expected


def test_symmetry_group_O():
    result = get_symmetry_ranges(symmetry_group="O")
    expected = SymmetryRanges(
        phi_min=0.0,
        phi_max=90.0,
        theta_min=0.0,
        theta_max=54.7356,
        psi_min=0.0,
        psi_max=360.0,
    )
    assert result == expected


def test_symmetry_group_I():
    result = get_symmetry_ranges(symmetry_group="I")
    expected = SymmetryRanges(
        phi_min=0.0,
        phi_max=180.0,
        theta_min=0.0,
        theta_max=31.7,
        psi_min=0.0,
        psi_max=360.0,
    )
    assert result == expected


def test_invalid_symmetry_group():
    with pytest.raises(ValueError, match="Symmetry group not recognized"):
        get_symmetry_ranges(symmetry_group="invalid")



"""Generate uniform 3D euler angles (ZYZ)."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("torch-so3")
except PackageNotFoundError:
    __version__ = "uninstalled"
__author__ = "Josh Dickerson"
__email__ = "jdickerson@berkeley.edu"

from .angular_ranges import SymmetryRanges, get_symmetry_ranges
from .base_s2_grid import cartesian_base_grid, healpix_base_grid, uniform_base_grid
from .local_so3_sampling import get_local_high_resolution_angles, get_roll_angles
from .uniform_so3_sampling import get_uniform_euler_angles

__all__ = [
    "get_uniform_euler_angles",
    "get_local_high_resolution_angles",
    "get_roll_angles",
    "get_symmetry_ranges",
    "SymmetryRanges",
    "uniform_base_grid",
    "healpix_base_grid",
    "cartesian_base_grid",
]



"""Functions for generating a base grid on the S^2 unit sphere."""

import platform
import warnings

import numpy as np
import torch
if platform.system() == "Windows":
    warnings.warn("healpy cannot be installed on Windows systems.", stacklevel=2)
else:
    import healpy as hp


def uniform_base_grid(
    theta_step: float = 2.5,
    theta_min: float = 0.0,
    theta_max: float = 180.0,
    phi_min: float = 0.0,
    phi_max: float = 360.0,
) -> torch.Tensor:
    """Generate a uniform base grid on the S^2 sphere.

    phi step is calculated by the position on the sphere (sin(theta))


    Parameters
    ----------
    theta_step : float, optional
        Angular step for theta in degrees. Default is 2.5
    theta_min : float, optional
        Minimum value for theta in degrees. Default is 0.0.
    theta_max : float, optional
        Maximum value for theta in degrees. Default is 180.0.
    phi_min : float, optional
        Minimum value for phi in degrees. Default is 0.0.
    phi_max : float, optional
        Maximum value for phi in degrees. Default is 360.0.

    Returns
    -------
    torch.Tensor
        Tensor of shape (N, 2) containing theta and phi values in degrees, where N is
        the number of angles pairs generated.
    """
    theta_step_rad = np.deg2rad(theta_step)
    phi_min_rad = np.deg2rad(phi_min)
    phi_max_rad = np.deg2rad(phi_max)

    # generate uniform set of theta values
    theta_all = np.arange(
        theta_min, theta_max + theta_step, theta_step, dtype=np.float64
    )
    theta_all = np.deg2rad(theta_all)

    # Phi step increment is modulated by the position on the sphere (sin(theta)), but
    # don't allow it to exceed the maximum step size
    phi_max_step_rad = phi_max_rad - phi_min_rad
    phi_step_all = np.abs(theta_step_rad / (np.sin(theta_all) + 1e-6))
    phi_step_all = np.clip(phi_step_all, a_min=None, a_max=phi_max_step_rad)
    if phi_max_step_rad > 0.0:
        phi_step_all = phi_max_step_rad / np.round(phi_max_step_rad / phi_step_all)
    else:
        phi_step_all *= 0.0

    # Now generate the angle pairs
    angle_pairs = []
    for i, phi_step in enumerate(phi_step_all):
        if phi_min_rad >= phi_max_rad or phi_step <= 0.0:
            phi_values = np.array([phi_min_rad], dtype=np.float64)
        else:
            phi_values = np.arange(phi_min_rad, phi_max_rad, phi_step, dtype=np.float64)
        theta_values = np.full_like(phi_values, theta_all[i])
        angle_pairs.append(np.stack([phi_values, theta_values], axis=1))

    # Convert back to degrees
    angle_pairs = np.rad2deg(np.concatenate(angle_pairs))

    return torch.tensor(angle_pairs, dtype=torch.float64)


def healpix_base_grid(
    theta_step: float = 2.5,
    theta_min: float = 0.0,
    theta_max: float = 180.0,
    phi_min: float = 0.0,
    phi_max: float = 360.0,
) -> torch.Tensor:
    """Generate a base grid on the S^2 sphere using HEALPix.

    phi step is calculated by the position on the sphere

    Parameters
    ----------
    theta_step : float, optional
        Angular step for theta in degrees. Default is 2.5
    theta_min : float, optional
        Minimum value for theta in degrees. Default is 0.0.
    theta_max : float, optional
        Maximum value for theta in degrees. Default is 180.0.
    phi_min : float, optional
        Minimum value for phi in degrees. Default is 0.0.
    phi_max : float, optional
        Maximum value for phi in degrees. Default is 360.0.

    Returns
    -------
    torch.Tensor
        Tensor of shape (N, 2) containing theta and phi values in degrees, where N is
        the number of angles pairs generated.
    """
    if platform.system() == "Windows":
        raise ImportError("healpy cannot be installed on Windows systems.")

    theta_step_rad = np.deg2rad(theta_step)

    estimated_num_pixels = int(4 * np.pi / (theta_step_rad * theta_step_rad))

    # Find the next largest npix value for a valid healpix grid
    exact_num_pixels = 0
    nside = np.ceil(np.sqrt(estimated_num_pixels / 12))
    while nside < 36:
        exact_num_pixels = 12 * nside**2
        if exact_num_pixels >= estimated_num_pixels:
            break
        nside += 1

    # Check to make sure the break condition was met
    if exact_num_pixels < estimated_num_pixels:
        raise ValueError("No valid nside found")

    # Generate the base grid
    pixels = np.arange(exact_num_pixels).astype(np.int64)
    theta_values, phi_values = hp.pix2ang(int(nside), pixels)
    theta_values = torch.tensor(np.rad2deg(theta_values), dtype=torch.float64)
    phi_values = torch.tensor(np.rad2deg(phi_values), dtype=torch.float64)

    # Remove values outside the desired range
    valid_indices = (
        (theta_values >= theta_min)
        & (theta_values <= theta_max)
        & (phi_values >= phi_min)
        & (phi_values <= phi_max)
    )

    # NOTE: This is doing batched indexing to not exceed memory limits
    if not torch.all(valid_indices):
        theta_result = []
        phi_result = []
        valid_indices = torch.nonzero(valid_indices).squeeze()
        for i in range(0, len(valid_indices), 256):
            batch_indices = valid_indices[i : i + 256]
            theta_result.append(theta_values[batch_indices])
            phi_result.append(phi_values[batch_indices])

        theta_values = torch.cat(theta_result)
        phi_values = torch.cat(phi_result)

    angle_pairs = torch.stack([phi_values, theta_values], dim=1)

    return angle_pairs


def cartesian_base_grid(
    theta_step: float = 2.5,
    phi_step: float = 1.5,
    theta_min: float = 0.0,
    theta_max: float = 180.0,
    phi_min: float = 0.0,
    phi_max: float = 360.0,
) -> torch.Tensor:
    """Generate a base grid on the S^2 sphere using a cartesian grid.

    phi step is now set explicitly. This will oversample the poles.

    Parameters
    ----------
    theta_step : float, optional
        Angular step for theta in degrees. Default is 2.5
    phi_step : float, optional
        Angular step for phi in degrees. Default is 1.5 degrees.
    theta_min : float, optional
        Minimum value for theta in degrees. Default is 0.0.
    theta_max : float, optional
        Maximum value for theta in degrees. Default is 180.0.
    phi_min : float, optional
        Minimum value for phi in degrees. Default is 0.0.
    phi_max : float, optional
        Maximum value for phi in degrees. Default is 360.0.

    Returns
    -------
    torch.Tensor
        Tensor of shape (N, 2) containing theta and phi values in degrees, where N is
        the number of angles pairs generated.
    """
    # Generate the base grid
    phi_values = torch.arange(
        phi_min,
        phi_max,
        phi_step,
        dtype=torch.float64,
    )

    theta_values = torch.arange(
        theta_min,
        theta_max,
        theta_step,
        dtype=torch.float64,
    )

    grid = torch.meshgrid(phi_values, theta_values, indexing="ij")
    euler_angles = torch.stack(grid, dim=-1).reshape(-1, 2)

    return euler_angles



"""Generates a set of Euler angles that uniformly samples SO(3) using Hopf fibration."""

import warnings
from typing import Literal, Optional

import torch

from torch_so3.base_s2_grid import (
    cartesian_base_grid,
    healpix_base_grid,
    uniform_base_grid,
)


def get_uniform_euler_angles(
    psi_step: float = 1.5,
    theta_step: float = 2.5,
    phi_step: Optional[float] = None,
    phi_min: float = 0.0,
    phi_max: float = 360.0,
    theta_min: float = 0.0,
    theta_max: float = 180.0,
    psi_min: float = 0.0,
    psi_max: float = 360.0,
    base_grid_method: Literal["uniform", "healpix", "cartesian"] = "uniform",
) -> torch.Tensor:
    """Generate sets of uniform Euler angles (ZYZ) using Hopf fibration.

    Parameters
    ----------
    psi_step: float, optional
        Angular step for psi in degrees. Default is 1.5 degrees.
    theta_step: float, optional
        Angular step for theta in degrees. Default is 2.5
        degrees.
    phi_step: float, optional
        Angular step for phi rotation in degrees. Only used when base_grid_method is
        "cartesian". Default is 2.5 degrees.
    phi_min: float, optional
        Minimum value for phi in degrees. Default is 0.0.
    phi_max: float, optional
        Maximum value for phi in degrees. Default is 360.0.
    theta_min: float, optional
        Minimum value for theta in degrees. Default is 0.0.
    theta_max: float, optional
        Maximum value for theta in degrees. Default is 180.0.
    psi_min: float, optional
        Minimum value for psi in degrees. Default is 0.0.
    psi_max: float, optional
        Maximum value for psi in degrees. Default is 360.0.
    base_grid_method: str, optional
        String literal specifying the method to generate the base grid. Default is
        "uniform". Options are "uniform", "healpix", and "cartesian".

    Returns
    -------
    torch.Tensor
        Tensor of shape (N, 3) containing Euler angles in degrees, where N is the
        number of angles generated.
    """
    # TODO: Validation of inputs, wrapping between zero and 2*pi, etc.

    if base_grid_method == "cartesian":
        # Handle cartesian_base_grid separately since it has a different signature
        actual_phi_step = 2.5 if phi_step is None else phi_step
        base_grid = cartesian_base_grid(
            theta_step=theta_step,
            phi_step=actual_phi_step,
            theta_min=theta_min,
            theta_max=theta_max,
            phi_min=phi_min,
            phi_max=phi_max,
        )
    else:
        # Handle uniform and healpix grids
        if base_grid_method == "uniform":
            base_grid_mth = uniform_base_grid
        elif base_grid_method == "healpix":
            base_grid_mth = healpix_base_grid
        else:
            raise ValueError(f"Invalid base grid method {base_grid_method}.")

        # Check if phi_step was specified for non-cartesian methods
        if phi_step is not None:
            warnings.warn(
                f"phi_step is being ignored for {base_grid_method} method.",
                stacklevel=2,
            )

        base_grid = base_grid_mth(
            theta_step=theta_step,
            theta_min=theta_min,
            theta_max=theta_max,
            phi_min=phi_min,
            phi_max=phi_max,
        )

    # Mesh-grid-like operation to include the in-plane rotation
    if psi_min >= psi_max:
        psi_all = torch.tensor([psi_min], dtype=torch.float64)
    else:
        psi_all = torch.arange(psi_min, psi_max, psi_step, dtype=torch.float64)

    psi_mesh = psi_all.repeat_interleave(base_grid.size(0))
    base_grid = base_grid.repeat(psi_all.size(0), 1)

    # Ordering of angles is (phi, theta, psi) for ZYZ intrinsic rotations
    # psi is the in-plane rotation
    all_angles = torch.cat([base_grid, psi_mesh[:, None]], dim=1)

    return all_angles



"""Generate finer angular search around multiple selected Euler angles."""

from typing import Literal

import torch

from torch_so3.uniform_so3_sampling import get_uniform_euler_angles

EPS = 1e-6


def get_local_high_resolution_angles(
    coarse_phi_step: float = 2.5,
    coarse_theta_step: float = 2.5,
    coarse_psi_step: float = 1.5,
    fine_phi_step: float = 0.1,
    fine_theta_step: float = 0.1,
    fine_psi_step: float = 0.1,
    base_grid_method: Literal["uniform", "healpix", "cartesian"] = "uniform",
) -> torch.Tensor:
    """Local orientation refinement from a coarse to fine grid.

    This function is essentially creating a high-resolution grid of Euler angles around
    the pole to sample all points within the coarse grid. It creates a "disk" of angles
    for the S^2 (theta, phi) sphere and also samples the in-plane angle (phi) uniformly.

    Parameters
    ----------
    coarse_phi_step : float
        Coarse step size for phi in degrees.
    coarse_theta_step : float
        Coarse step size for theta in degrees.
    coarse_psi_step : float
        Coarse step size for psi in degrees.
    fine_phi_step : float
        Finer step size for phi in degrees.
    fine_theta_step : float
        Finer step size for theta in degrees.
    fine_psi_step : float
        Finer step size for psi in degrees.
    base_grid_method : Literal["uniform", "healpix", "cartesian"]
        Method to generate the base grid.

    Returns
    -------
    euler_angles : torch.Tensor
        Tensor of shape (N, 3) containing Euler angles in degrees, where N is the
        number of angles generated. Angles exist around pole (0, 0, 0) and define grid
        to search over.
    """
    if base_grid_method == "cartesian":
        euler_angles = get_uniform_euler_angles(
            phi_step=fine_phi_step,
            theta_step=fine_theta_step,
            psi_step=fine_psi_step,
            phi_min=-coarse_phi_step,  # Completely sample for uniform base grid?
            phi_max=coarse_phi_step,
            theta_min=-coarse_theta_step,
            theta_max=coarse_theta_step,
            psi_min=-coarse_psi_step,
            psi_max=coarse_psi_step + EPS,
            base_grid_method=base_grid_method,
        )
    else:
        euler_angles = get_uniform_euler_angles(
            psi_step=fine_psi_step,
            theta_step=fine_theta_step,
            phi_min=-coarse_psi_step,  # Completely sample for uniform base grid?
            phi_max=coarse_psi_step,
            theta_min=-coarse_theta_step,
            theta_max=coarse_theta_step,
            psi_min=-coarse_psi_step,
            psi_max=coarse_psi_step + EPS,
            base_grid_method=base_grid_method,
        )

    return euler_angles


def get_roll_angles(
    psi_step: float = 0.5,
    psi_min: float = -10.0,
    psi_max: float = 10.0,
    theta_step: float = 0.5,
    theta_min: float = -10.0,
    theta_max: float = 10.0,
    roll_axis: torch.Tensor = None,
    roll_axis_step: float = 2.0,
) -> torch.Tensor:
    """Generate a grid of Euler angles for rotations around Z and a roll axis.

    This function generates ZYZ Euler angles (in degrees) that represent:
    1. Rotations around the Z axis (primary rotation)
    2. Rotations around a "roll axis" that's in the XY plane
       (orthogonal to Z but not necessarily aligned with Y)

    If no specific roll axis is provided, it samples multiple possible roll axes
    by varying angles in the XY plane and using complementary rotations that
    sum to the desired psi angle.

    Parameters
    ----------
    psi_step : float, optional
        Step size for the in-plane rotation (psi) in degrees, by default 0.5
    psi_min : float, optional
        Minimum value for psi in degrees, by default -10.0
    psi_max : float, optional
        Maximum value for psi in degrees, by default 10.0
    theta_step : float, optional
        Step size for the roll rotation (theta) in degrees, by default 0.5
    theta_min : float, optional
        Minimum value for theta in degrees, by default -10.0
    theta_max : float, optional
        Maximum value for theta in degrees, by default 10.0
    roll_axis : torch.Tensor, optional
        If provided, specifies a fixed roll axis as [x, y] in Cartesian coordinates.
        If None, searches through multiple roll axes, by default None.
    roll_axis_step : float, optional
        Angular step for searching different roll axes in degrees, by default 2.0

    Returns
    -------
    torch.Tensor
        Tensor of shape (N, 3) containing Euler angles (phi, theta, psi) in degrees
        representing the sampled rotations.
    """
    # Generate arrays for psi (rotation around Z) and theta (roll angle)
    psi_values = torch.arange(psi_min, psi_max + EPS, psi_step, dtype=torch.float32)
    theta_values = torch.arange(
        theta_min, theta_max + EPS, theta_step, dtype=torch.float32
    )

    # Set up the roll axis search
    if roll_axis is None:
        # Search for roll axes by sampling angles in the XY plane
        roll_angles = torch.arange(0, 180, roll_axis_step, dtype=torch.float32)
    else:
        # Use the given roll axis [x,y] to determine the roll angle
        # Convert from Cartesian coordinates to angle in the XY plane
        roll_angle = torch.rad2deg(torch.atan2(roll_axis[1], roll_axis[0]))
        # Normalize to 0-180 range
        if roll_angle < 0:
            roll_angle += 180
        roll_angles = torch.tensor([roll_angle], dtype=torch.float32)

    # Create meshgrid for all dimensions
    roll_grid, theta_grid, psi_grid = torch.meshgrid(
        roll_angles, theta_values, psi_values, indexing="ij"
    )

    # Flatten all grids
    roll_flat = roll_grid.reshape(-1)
    theta_flat = theta_grid.reshape(-1)
    psi_flat = psi_grid.reshape(-1)

    # For each combination (I do extrinsic in my head so go backwards):
    # 1. First Euler angle (phi): Rotate to align back with the psi_angle
    # 2. Second Euler angle (theta): Apply theta rotation around roll axis
    # 3. Third Euler angle (psi): Rotate to align with roll axis

    # Second Euler angle: theta rotation around roll axis
    theta = theta_flat

    # Third Euler angle: rotate back to achieve the desired psi rotation
    # We rotate back by (roll_angle - psi_value) to get a net psi rotation
    psi = 360 - roll_flat

    # Adjust back-rotation to achieve desired rotation axis angle
    phi = roll_flat + psi_flat

    # Ensure phi is within [0, 360) range
    phi = phi % 360

    # Stack the Euler angles
    all_angles = torch.stack([phi, theta, psi], dim=1)

    return all_angles



"""Generate Euler angles ranges applying symmetry operations."""

from typing import NamedTuple


class SymmetryRanges(NamedTuple):
    """NamedTuple child class for 'get_symmetry_ranges' return type."""

    phi_min: float
    phi_max: float
    theta_min: float
    theta_max: float
    psi_min: float
    psi_max: float


def get_symmetry_ranges(
    symmetry_group: str = "C",
    symmetry_order: int = 1,
) -> SymmetryRanges:
    """
    Generate Euler angle ranges based on symmetry group.

    Parameters
    ----------
    symmetry_group : str, optional
        Symmetry group of the particle (C, D, T, O, I). Default is "C".
    symmetry_order : int, optional
        Order of the symmetry group. Default is 1.

    Returns
    -------
    SymmetryRanges
        NamedTuple containing the min/max for phi, theta, and psi.
    """
    # Convert to upper case
    symmetry_group = symmetry_group.upper()
    phi_max = 360.0
    theta_max = 180.0
    psi_max = 360.0
    if symmetry_group == "C":
        phi_max = 360 / float(symmetry_order)
    elif symmetry_group == "D":
        phi_max = 360 / float(symmetry_order)
        theta_max = 90.0
    elif symmetry_group == "T":
        phi_max = 180.0
        theta_max = 54.7356
    elif symmetry_group == "O":
        phi_max = 90.0
        theta_max = 54.7356
    elif symmetry_group == "I":
        phi_max = 180.0
        theta_max = 31.7
    else:
        raise ValueError("Symmetry group not recognized")

    return SymmetryRanges(
        phi_min=0.0,
        phi_max=phi_max,
        theta_min=0.0,
        theta_max=theta_max,
        psi_min=0.0,
        psi_max=psi_max,
    )



"""Tests for `torch_angular_search` package."""

import platform

import numpy as np
import pytest

from torch_so3.local_so3_sampling import (
    get_local_high_resolution_angles,
    get_roll_angles,
)
from torch_so3.uniform_so3_sampling import get_uniform_euler_angles

# TODO: Check actual values of returned tensors


def test_get_uniform_euler_angles():
    # Test the angle generator
    angles = get_uniform_euler_angles(base_grid_method="uniform")
    assert angles.shape == (1584480, 3)

    # Ensure that the angles are within the desired (default) range
    assert (angles[:, 0] >= 0).all()
    assert (angles[:, 0] <= 360).all()
    assert (angles[:, 1] >= 0).all()
    assert (angles[:, 1] <= 180).all()
    assert (angles[:, 2] >= 0).all()
    assert (angles[:, 2] <= 360).all()


@pytest.mark.skipif(
    platform.system() == "Windows", reason="healpy is not supported on Windows"
)
def test_get_uniform_euler_angles_healpix():
    angles = get_uniform_euler_angles(base_grid_method="healpix")
    assert angles.shape == (1658880, 3)


def test_get_local_high_resolution_angles():
    local_angles = get_local_high_resolution_angles()
    assert local_angles.shape == (1581, 3)

    # range tests for angles
    assert (local_angles[:, 0] >= -1.51).all()
    assert (local_angles[:, 0] <= 1.51).all()
    assert (local_angles[:, 1] >= -2.51).all()
    assert (local_angles[:, 1] <= 2.51).all()
    assert np.allclose(local_angles[:, 2].min().item(), -1.50)
    assert np.allclose(local_angles[:, 2].max().item(), 1.50)


def test_get_roll_angles():
    roll_angles = get_roll_angles()
    assert roll_angles.shape == (151290, 3)

    # range tests for angles
    assert (roll_angles[:, 1] >= -10.01).all()
    assert (roll_angles[:, 1] <= 10.01).all()
